/* Copyright 2024 Jan-Michael Brummer
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "msg-json-utils.h"

const char *
msg_json_object_get_string (JsonObject *object,
                            const char *name)
{
  if (json_object_has_member (object, name))
    return json_object_get_string_member (object, name);

  return NULL;
}

gboolean
msg_json_object_get_boolean (JsonObject *object,
                             const char *name)
{
  if (json_object_has_member (object, name)) {
    return json_object_get_boolean_member (object, name);
  }

  return FALSE;
}
